#ifndef _EDITWING_DOC_H_
#define _EDITWING_DOC_H_
#include "ewCommon.h"
#ifndef __ccdoc__
namespace editwing {
namespace doc {
#endif



class DocImpl;
class DocEvHandler;
class Command;
class Insert;
class Delete;
class Replace;



//=========================================================================
//@{ @pkg editwing.Doc //@}
//@{
//	f[^
//
//	̃NX͒PȂC^[tFCXŁA
//	class DocImpl ōsB̂ŁAڂ͂QƂ̂ƁB
//@}
//=========================================================================

class Document : public ki::Object
{
public:

	//@{ ȂRXgN^ //@}
	Document();
	~Document();

	//@{ t@CJ //@}
	void OpenFile( ki::aptr<ki::TextFileR> tf );

	//@{ t@Cۑ //@}
	void SaveFile( ki::TextFileW& tf );

	//@{ ej //@}
	void ClearAll();

	//@{ R}hs //@}
	void Execute( const Command& cmd );

	//@{ AhD //@]
	void Undo();

	//@{ hD //@]
	void Redo();

	//@{ AhD񐔐 //@]
	void SetUndoLimit( long lim );

	//@{ ύXtONA //@}
	void ClearModifyFlag();

	//@{ Cxgnho^ //@}
	void AddHandler( DocEvHandler* eh );

	//@{ Cxgnh //@}
	void DelHandler( DocEvHandler* eh );

	//@{ L[[h`؂ւ //@}
	void SetKeyword( const unicode* defbuf, ulong siz=0 );

public:

	//@{ NX //@}
	DocImpl& impl() { return *impl_; }

	//@{ s //@}
	ulong tln() const;

	//@{ sobt@ //@}
	const unicode* tl( ulong i ) const;

	//@{ s //@}
	ulong len( ulong i ) const;

	//@{ w͈͂̃eLXg̒ //@}
	ulong getRangeLength( const DPos& stt, const DPos& end ) const;

	//@{ w͈͂̃eLXg //@}
	void getText( unicode* buf, const DPos& stt, const DPos& end ) const;

	//@{ AhD\H //@}
	bool isUndoAble() const;

	//@{ hD\H //@}
	bool isRedoAble() const;

	//@{ ύXς݁H //@}
	bool isModified() const;

	//@{ rW[tOi}NR}hŝݐj //@}
	void setBusyFlag( bool b=true ) { busy_ = b; }
	bool isBusy() const { return busy_; }

private:

	// 
	ki::dptr<DocImpl> impl_;
	bool busy_;

private:

	NOCOPY(Document);
};



//=========================================================================
//@{
//	CxgnhC^[tFCX
//
//	hLg甭Cxgi}/폜ȂǂȂǁcj
//	󂯎肽ꍇ́ÃC^[tFCXpAKXnh
//	ƁBView̍ĕ`揈ȂǂʂĎsĂB
//@}
//=========================================================================

class DocEvHandler
{
public:
	//@{
	//	eLXgeύXꂽƂɔ
	//	@param s        ύX͈͂̐擪
	//	@param e        ύX͈͂̏I[(O)
	//	@param e2       ύX͈͂̏I[()
	//	@param reparsed e2̃RgAEgԂωĂtrue
	//	@param nmlcmd   }/폜/uȂtrueAt@CJ/SuȂfalse
	//@}
	virtual void on_text_update( const DPos& s,
		const DPos& e, const DPos& e2, bool reparsed, bool nmlcmd ) {}

	//@{
	//	L[[hύXꂽƂɔ
	//@}
	virtual void on_keyword_change() {}

	//@{
	//	_[eBtOύXꂽƂɔ
	//@}
	virtual void on_dirtyflag_change( bool dirty ) {}
};



//=========================================================================
//@{
//	R}hC^[tFCX
//
//	hLǵACommand hNX̃CX^X
//	operator() ĂяoƂŁAFXȑsBƂ肠
//	̓Iɂ Insert/Delete/Replace ̂RBƂŃ}NR}hp
//	NX肾ǁAƂ肠͕ۗB
//@}
//=========================================================================

class Command : public ki::Object
{
protected:
	friend class UnReDoChain;
	friend class MacroCommand;
	virtual Command* operator()( Document& doc ) const = 0;
};



//=========================================================================
//@{
//	}R}h
//@}
//=========================================================================

class Insert : public Command
{
public:

	//@{
	//	@param s }ʒu
	//	@param str }
	//	@param len ̒
	//	@param del R}hIdelete [] strĂ悢H
	//@}
	Insert( const DPos& s, const unicode* str, ulong len, bool del=false );
	~Insert();

private:

	Command* operator()( Document& doc ) const;
	DPos           stt_;
	const unicode* buf_;
	ulong          len_;
	bool           del_;
};



//=========================================================================
//@{
//	폜R}h
//@}
//=========================================================================

class Delete : public Command
{
public:

	//@{
	//	@param s Jnʒu
	//	@param e I[ʒu
	//@}
	Delete( const DPos& s, const DPos& e );

private:

	Command* operator()( Document& doc ) const;
	DPos stt_;
	DPos end_;
};




//=========================================================================
//@{
//	uR}h
//@}
//=========================================================================

class Replace : public Command
{
public:

	//@{
	//	@param s Jnʒu
	//	@param e I[ʒu
	//	@param str }
	//	@param len ̒
	//	@param del R}hIdelete [] strĂ悢H
	//@}
	Replace( const DPos& s, const DPos& e,
		const unicode* str, ulong len, bool del=false );
	~Replace();

private:

	Command* operator()( Document& doc ) const;
	DPos           stt_;
	DPos           end_;
	const unicode* buf_;
	ulong          len_;
	bool           del_;
};



//=========================================================================
//@{
//	}NR}h
//
//	̃R}h̃R}hƂĘAsB
//	AInsert/Delete/ReplacesтɓR
//	̈ʒu͕ω̂AɊւϊ
//	sȂBȂ킿AInsert->Delete->Insert ݂
//	AƂ́As╶̕ωlȂ
//	l߂ĂƂKvɂȂB̂ŁA܂gȂ(^^;
//@}
//=========================================================================

class MacroCommand : public Command
{
public:
	//@{ R}h̒ǉ //@}
	void Add( Command* cmd ) { arr_.Add(cmd); }

	//@{ R}h //@}
	ulong size() const { return arr_.size(); }

	//@ fXgN^ //@}
	~MacroCommand()
	{
		for( ulong i=0,e=arr_.size(); i<e; ++i )
			delete arr_[i];
	}

private:
	Command* operator()( Document& doc ) const;
	ki::storage<Command*> arr_;
};



//=========================================================================

}}     // namespace editwing::document
#endif // _EDITWING_DOC_H_
